#lang racket

; Atentie! Trebuie ales de jos din DrRacket: Determine language from source.

; 1. (2p)
; Sa se scrie o functie care implementeaza quicksort. Functia va primi
; ca parametru o functie comparator. Folositi o
; constructie de tip let pentru a scrie cat mai putin cod.

; exemplu de comparator:
(define comp-less-than
  (lambda (a b)
    (< a b)))

(define qsort
  (lambda (numbers comp)
    (if (null? numbers)
        numbers
        (let* ((pivot (car numbers))
               (rest (cdr numbers))
               (smaller (filter (lambda (x) (comp x pivot)) rest))
               (greater (filter (lambda (x) (not (comp x pivot))) rest)))
          (append (qsort smaller comp) (list pivot) (qsort greater comp))))))

; puteti testa folosind:
(qsort '(5 4 3 2 1) comp-less-than)


; 2. (2p)
; Sa se scrie o functie care gaseste radacina unei functii date.
; Mai multe informatii la http://en.wikipedia.org/wiki/Bisection_method

; Se dau urmatoarele:
; o functie, f1 -- f1(x) = x + 5
; un numar maxim de iteratii (MAX_ITER)
; o valoare a preciziei cerute (EPS)
(define f1
  (lambda (x)
    (+ x 5)))

(define MAX_ITER 100)
(define EPS 0.001)

; Veti folosi un named let pentru a simula iteratia.
(define bisection
  (lambda (f interval)
    (let bisect ((left (car interval)) (right (cdr interval)) (iter 0))
      (let* ((mid (/ (+ left right) 2))
             (val-mid (f mid))
             (val-left (f left))
             (val-right (f right)))
        (if (or (= val-mid 0) (<= (- right left) EPS) (= iter MAX_ITER))
            mid
            (if (> 0 (* val-left val-mid))
                (bisect left mid (+ 1 iter))
                (bisect mid right (+ 1 iter))))))))

(define interval  '(-10 . 10))

; testati folosind urmatoarele doua apeluri:
(bisection f1 interval)
(exact->inexact (bisection f1 interval))
; (exact->inexact se foloseste pentru a converti numere rationale in float
; puteti schimba valorile MAX_ITER si EPS pentru a experimenta


; cititi despre functii pe stringuri aici:
; http://docs.racket-lang.org/reference/strings.html

; 3. (2p)
; Sa se scrie o functie care ordoneaza o lista de string-uri astfel
; incat anagramele sa fie alaturate.
; hint: folositi functia qsort definita mai sus. Comparatorul pe care
; il veti folosi va trebui sa compare doua string-uri ordonate lexicografic

(define anagram-comparator
  (lambda (str1 str2)
    (string<? (list->string (qsort (string->list str1) char<?))
              (list->string (qsort (string->list str2) char<?)))))

(define sort-by-anagrams
  (lambda (strings)
    (qsort strings anagram-comparator)))

(sort-by-anagrams '("aba" "aca" "baa" "acc" "aac" "ccb" "cccd")) ; =>
; '("aba" "baa" "aca" "aac" "acc" "ccb" "cccd").


; 4. (2p)
; Fie urmatoarea codificare pentru nodurile unui arbore ce reprezinta un
; sistem de fisiere:
; (nume_director lista_continut) pt directoare
; nume_fisier pentru fisiere
; un director vid este codificat ca: (nume_director null)
; de exemplu: ("root/" (("gigi/" ("a.txt" "b.txt")))) reprezinta arborele
; pentru caile root/gigi/a.txt si root/gigi/b.txt

; Sa se scrie o functie care primeste ca parametri un astfel de arbore si
; un nume de fisier/director si intoarce calea catre el, daca exista.
; Daca fisierul/directorul nu exista, intoarce #f.
(define fs '("/"
             (("home/" (("gigi/" ("a.txt" "b.txt")) ("dorel/" ())))
              ("etc/" (("network/" ("interfaces")) "passwd" "shadow"))
              ("root/" ()))))

(define get-path-helper
  (lambda (tree name path)
    (if (null? tree)
        #f
        (if (list? tree)
            ; este director
            (if (equal? (car tree) name) ; am gasit ce cautam?
                (string-append path name)
                ; daca nu, trebuie cautat in subarbore
                (let loop ((remaining (cadr tree)))
                  (if (null? remaining)
                      #f ; nu mai am ce sa caut si nu am gasit
                      (let ((found (get-path-helper (car remaining) name (string-append path (car tree)))))
                        (if found
                            found ; am gasit ceva diferit de #f, deci calea cautata
                            (loop (cdr remaining))))))) ; nu am gasit, caut in restul
            ; este fisier
            (if (equal? name tree)
                (string-append path name)
                #f)))))

(define get-path
  (lambda (tree name)
    (get-path-helper tree name "")))

; testati folosind
(get-path fs "interfaces") ; => "/etc/network/interfaces"
(get-path fs "random") ; => #f


; 5. (2p)
; In contextul exercitiului anterior, sa se scrie o functie care primeste
; ca parametri un arbore si o cale si verifica daca acea cale este valida.

(define path->list
  (lambda (path)
    (regexp-split #px"/" path)))

(define dir-name
  (lambda (dir)
    (substring dir 0 (- (string-length dir) 1))))


(define check-path-helper
  (lambda (tree path)
    (if (null? path)
        #t
        (if (null? tree)
            #f
            (if (list? tree) ; daca este un director
                (if (equal? (car path) (dir-name (car tree))) ; face parte directorul din cale?
                    ; daca da, caut in continut (fold de "SAU" pe ce intoarce functia pe continut
                    (foldl (lambda (x y) (or x y)) #f (map (lambda (x) (check-path-helper x (cdr path))) (cadr tree)))
                    #f)
                (equal? tree (car path)))))))

(define check-path
  (lambda (tree path)
    (check-path-helper tree (path->list path))))

(check-path fs "/home/gigi/a.txt") ; => #t
(check-path fs "/proc") ; => #f


; Bonusul e bonus. :)
; 6. (2p)
; Se dau o structura de directoare sub forma de arbore si o lista de directoare si de fisiere. 
; Sa se construiasca toate caile posibile si sa se returneze o lista cu toate caile valide.
; Verificarea daca o cale este valida se face folosind functia check-path de la exercitiul anterior.
; Numele de directoare se termina cu "/"

;(valid-path-permutations fs '("home/" "a.txt" "dorel/")) ;=> '("home/" "home/dorel/")

